import csv
import argparse
import concurrent.futures
from dataclasses import dataclass
import os
from typing import Literal, Optional, cast
from PIL import Image

from .style import Style
from .mkpdf import mkpdf  # Import the mkpdf function
from .scryfall_api import get_card  # Import the makecard function
from .datatypes import ArtInfo
from functools import partial
from .cache_url import get_cache


@dataclass
class GenerationResult:
    rownr: int
    copies: int
    card: list[str] | list[Image.Image]


@dataclass
class XMLCard:
    id: str
    slots: list[int]

    def to_xml(self) -> str:
        return f"""        <card>
            <id>{os.path.abspath(self.id)}</id>
            <slots>{",".join([str(slot) for slot in self.slots])}</slots>
        </card>"""


def generate_card_images(
    style: Style,
    output_path: str,
    mode: Literal["pdf", "folder", "xml"],
    cache: Optional[str],
    firstprint: bool,
    input: tuple[int, list[str]],
) -> GenerationResult:
    """Helper function to generate images for a single row using makecard."""
    get_cache(cache, True)
    index, row = input
    amount = int(row[0])  # Number of copies of the card
    query = row[1] + (" is:firstprint" if firstprint else "")
    art1_url = row[2] if row[2] else None  # URL for first artwork, if provided
    art2_url = row[3] if row[3] else None  # URL for second artwork, if provided
    art1_xoffset = float(row[6]) if row[6] else None
    art1_yoffset = float(row[7]) if row[7] else None
    art1_zoom = float(row[8]) if row[8] else None
    art2_xoffset = float(row[9]) if row[9] else None
    art2_yoffset = float(row[10]) if row[10] else None
    art2_zoom = float(row[11]) if row[11] else None

    art1 = ArtInfo(
        url=art1_url,
        x_perc_offset=art1_xoffset,
        y_perc_offset=art1_yoffset,
        zoom=art1_zoom,
    )
    art2 = ArtInfo(
        url=art2_url,
        x_perc_offset=art2_xoffset,
        y_perc_offset=art2_yoffset,
        zoom=art2_zoom,
    )

    style = (
        Style.load(row[4], style.add_bleed) if row[4] else style
    )  # Resource pack, default if blank

    # Call makecard once to generate the card images
    # print(f"Generating images for: {query}")
    try:
        print(f"Generating images for: {query}")
        try:
            card = get_card(query, art1, art2)
            print("Generated")
        except KeyError:
            from traceback import format_exc
            from sys import stderr

            print(format_exc(), file=stderr)
            print(f"Error: Card '{query}' not found.")
            exit(1)
        # if mode == "folder":
        #     filename = f"{output_path}/{card.set}{card.cn_int:03d}{card.cn_add}.png"
        #     if os.path.exists(filename):
        #         print(f"Skipping {card.set}{card.cn_int:03d}{card.cn_add}")
        #         return GenerationResult(index, amount, [filename])
        try:
            generated_images = style.render_card(card)
        except ValueError:
            from traceback import format_exc
            from sys import stderr

            print(format_exc(), file=stderr)
            print(f"Error: Card '{query}' could not be rendered.")
            exit(1)
    except Exception as e:
        print(f"Error generating images for '{query}': {e}")
        exit(1)

    # Duplicate images based on the amount specified
    # print(f"Generated {len(generated_images)} images for '{query}'")
    if mode == "pdf":
        return GenerationResult(index, amount, generated_images)
    else:
        print(f"Saving {card.faces[0].name} to {output_path}")
        files = []
        if len(generated_images) > 1:
            for i, image in enumerate(generated_images):
                image.save(
                    f"{output_path}/{card.set}{card.cn_int:03d}{card.cn_add}_{i}.png"
                )
                files.append(
                    f"{output_path}/{card.set}{card.cn_int:03d}{card.cn_add}_{i}.png"
                )
        else:
            generated_images[0].save(
                f"{output_path}/{card.set}{card.cn_int:03d}{card.cn_add}.png"
            )
            files.append(f"{output_path}/{card.set}{card.cn_int:03d}{card.cn_add}.png")

        return GenerationResult(index, amount, files)


def process_csv_and_generate_output(
    csv_file_path: str,
    style: Style,
    mode: Literal["pdf", "folder", "xml"],
    output_path: str = "output",
    num_processes: Optional[int] = None,
    cache: Optional[str] = None,
    firstprint: bool = True,
) -> None:
    """Reads the CSV file, generates card images in parallel, and saves them as a PDF."""

    MAX_COLS = 12
    all_images = []

    # Read the CSV file and prepare rows for processing
    with open(csv_file_path, mode="r", newline="", encoding="utf-8") as csvfile:
        reader = csv.reader(csvfile, delimiter=";")
        rows = list(reader)  # Collect rows for processing
        for row in rows:
            if len(row) < MAX_COLS:
                row.extend(
                    [""] * (MAX_COLS - len(row))
                )  # Fill missing columns with empty strings

    # Use ProcessPoolExecutor to parallelize makecard calls with configurable max_workers
    if mode == "folder":
        os.makedirs(output_path, exist_ok=True)

    with concurrent.futures.ThreadPoolExecutor(max_workers=num_processes) as executor:
        # Map each row to generate_card_images function in parallel
        results = list(
            executor.map(
                partial(
                    generate_card_images,
                    style,
                    output_path,
                    mode,
                    cache,
                    firstprint,
                ),
                enumerate(rows),
            )
        )
    # for row in rows:
    #     results.append(generate_card_images(style, row))

    # Flatten the list of lists of images
    for result in results:
        if len(result.card) > 0 and isinstance(result.card[0], Image.Image):
            all_images.extend(result.card * result.copies)

    # print(f"Generated {len(all_images)} card images.")

    if mode == "pdf":
        # Generate the PDF from the collected images
        print("Generating PDF file.")
        mkpdf(all_images, output_path)
    elif mode == "folder":
        print("Generating XML file.")
        quantity = sum([result.copies for result in results])
        all_bracket = [
            18,
            36,
            55,
            72,
            90,
            108,
            126,
            144,
            162,
            180,
            198,
            216,
            234,
            396,
            504,
            612,
        ]
        best_bracket = [i for i in all_bracket if i >= quantity][0]

        start = f"""<order>
    <details>
        <quantity>{quantity}</quantity>
        <bracket>{best_bracket}</bracket>
        <stock>(S30) Standard Smooth</stock>
        <foil>false</foil>
    </details>
"""

        front_cards = []
        back_cards = []
        next_slot = 0
        for card in results:
            slots = list(range(next_slot, next_slot + card.copies))
            next_slot += card.copies
            card.card = cast(list[str], card.card)
            front_cards.append(XMLCard(card.card[0], slots))
            if len(card.card) > 1:
                back_cards.append(XMLCard(card.card[1], slots))

        fronts = (
            "    <fronts>\n"
            + "\n".join([card.to_xml() for card in front_cards])
            + "\n    </fronts>"
        )
        if len(back_cards) > 0:
            backs = (
                "    <backs>\n"
                + "\n".join([card.to_xml() for card in back_cards])
                + "\n    </backs>"
            )
        else:
            backs = ""

        end = "\n    <cardback>1LrVX0pUcye9n_0RtaDNVl2xPrQgn7CYf</cardback>\n</order>"

        xml = start + fronts + backs + end
        with open(output_path + "/deck.xml", "w") as f:
            f.write(xml)


def main():
    # Set up argument parser for command-line arguments
    parser = argparse.ArgumentParser(
        description="Generate a PDF of card images from a CSV file."
    )
    parser.add_argument("csv_file", help="Path to the input CSV file.")
    parser.add_argument(
        "--pdf", action="store_true", help="Generate PDF file.", default=False
    )
    parser.add_argument(
        "-o",
        "--output",
        default="output",
        help="Path to the output PDF file or folder (default: output).",
    )
    parser.add_argument(
        "-p",
        "--processes",
        type=int,
        default=None,
        help="Number of parallel processes to use (default: number of CPU cores).",
    )

    parser.add_argument("-c", "--cache", default=None)

    # Parse the arguments
    args = parser.parse_args()

    # Call the main processing function with parsed arguments
    process_csv_and_generate_output(
        args.csv_file, args.pdf, args.output, args.processes, args.cache
    )


if __name__ == "__main__":
    main()
