from __future__ import annotations

from argparse import ArgumentParser
import json

from .style import Style
from .datatypes import ArtInfo
from .scryfall_api import get_card
from .cache_url import get_cache
from .deck import process_csv_and_generate_output


def main():
    # card = get_card("Murktide Regent")
    parser = ArgumentParser()
    parser.add_argument("--resource-pack", "-R", default="resource_pack.zip")
    parser.add_argument("--cache", "-c", default=None)
    parser.add_argument("--add_bleed", "-b", action="store_true")

    modeparser = parser.add_subparsers(dest="mode", required=True)
    cardparser = modeparser.add_parser("card")

    cardparser.add_argument("--art1-url")
    cardparser.add_argument("--art2-url")
    cardparser.add_argument("--art1-x", type=float, default=None)
    cardparser.add_argument("--art1-y", type=float, default=None)
    cardparser.add_argument("--art1-zoom", type=float, default=None)
    cardparser.add_argument("--art2-x", type=float, default=None)
    cardparser.add_argument("--art2-y", type=float, default=None)
    cardparser.add_argument("--art2-zoom", type=float, default=None)
    output_group = cardparser.add_mutually_exclusive_group(required=True)
    output_group.add_argument("--save", "-s", action="store_true")
    output_group.add_argument("--display", "-d", action="store_true")
    input_group = cardparser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("--query", "-q")
    input_group.add_argument("--json", "-j")

    deckparser = modeparser.add_parser("deck")
    deckparser.add_argument("--firstprint", "-f", action="store_true")
    deckparser.add_argument("--process", "-p", type=int, default=None)
    deckparser.add_argument(
        "--type", "-t", type=str, choices=["pdf", "folder"], default="folder"
    )
    deckparser.add_argument("--output", "-o", type=str, required=True)
    deckparser.add_argument("csv_file")

    args = parser.parse_args()
    style = Style.load(args.resource_pack, add_bleed=args.add_bleed)

    if args.mode == "card":
        querystr = args.query
        query = args.query
        if args.json:
            querystr = args.json
            with open(args.json) as f:
                query = json.load(f)

        cache = get_cache(args.cache)
        art1 = ArtInfo(
            args.art1_url,
            x_perc_offset=args.art1_x,
            y_perc_offset=args.art1_y,
            zoom=args.art1_zoom,
        )
        art2 = ArtInfo(
            args.art2_url,
            x_perc_offset=args.art2_x,
            y_perc_offset=args.art2_y,
            zoom=args.art2_zoom,
        )
        try:
            card = get_card(query, art1=art1, art2=art2)
        except KeyError:
            # import traceback
            from traceback import format_exc
            from sys import stderr

            print(format_exc(), file=stderr)
            print(f"Could not render card {querystr}", file=stderr)
            exit(1)
        images = style.render_card(card)

        for n, image in enumerate(images):
            if args.display:
                image.show()
            else:
                suffix = f"_{n}" if len(images) > 1 else ""
                filename = f"{card.set}{card.cn_int:03d}{card.cn_add}{suffix}.png"
                image.save(filename)
    elif args.mode == "deck":
        process_csv_and_generate_output(
            args.csv_file,
            style,
            args.type,
            args.output,
            args.process,
            args.cache,
            args.firstprint,
        )

    # old_frame = Style.load(args.resource_pack)
    #
    # cards = [
    #     get_card(query, art1=args.art1_url, art2=args.art2_url) for query in queries
    # ]
    # for card in cards:
    #     images = old_frame.render_card(card, args.resource_pack)
    #     for n, image in enumerate(images):
    #         if args.folder is not None:
    #             image.save(
    #                 f"{args.folder}/{card.faces[n].name}-{card.set}-{card.cn}.png"
    #             )
    #         else:
    #             image.show()


if __name__ == "__main__":
    main()
