from __future__ import annotations
from dataclasses import dataclass
from json import load
import sys
from typing import Any, Literal, Optional
from zipfile import ZipFile

from PIL import Image, ImageOps

from .scryfall_api import get_card
from .datatypes import ArtInfo, BBox, Card, Face
from .merge import merge, overlay


@dataclass
class Borders:
    lands: dict[str, Image.Image]
    saga: dict[str, Image.Image]
    default: dict[str, Image.Image]
    token: dict[str, Image.Image]
    token_textless: dict[str, Image.Image]
    backs: dict[str, Image.Image]

    @classmethod
    def load(cls, archive: ZipFile, stylefile: dict[str, Any]) -> Borders:
        card_styles = stylefile["dimensions"]
        borders = {card_style: {} for card_style in card_styles}
        lands = {}
        # saga = {}
        # default = {}
        backs = {}

        l_colors = ["w", "u", "b", "r", "g", "l"]
        l_files = ["box", "lines"]

        for color in l_colors:
            for file in l_files:
                with archive.open(f"borders/land/{color}_{file}.png") as imgdata:
                    image = Image.open(imgdata)
                    image.load()
                    lands[f"{color}_{file}"] = image
        with archive.open("borders/land/l_back.png") as l_back:
            lands["l_back"] = Image.open(l_back)
            lands["l_back"].load()

        s_colors = ["w", "u", "b", "r", "g", "l"]

        for color in s_colors:
            with archive.open(f"borders/saga/{color}_saga.png") as imgdata:
                borders["saga"][color] = Image.open(imgdata)
                borders["saga"][color].load()

        f_colors = ["w", "u", "b", "r", "g", "a", "c", "m"]

        with archive.open("borders/default/outline.png") as imgdata:
            borders["default"]["outline"] = Image.open(imgdata)
            borders["default"]["outline"].load()

        with archive.open("borders/default/leveler_shadow.png") as imgdata:
            borders["default"]["leveler_shadow"] = Image.open(imgdata)
            borders["default"]["leveler_shadow"].load()

        with archive.open("borders/token/l.png") as imgdata:
            borders["token"]["l"] = Image.open(imgdata)
            borders["token"]["l"].load()

        with archive.open("borders/token/l_textless.png") as imgdata:
            borders["token_textless"]["l"] = Image.open(imgdata)
            borders["token_textless"]["l"].load()

        for color in f_colors:
            try:
                with archive.open(f"borders/token/{color}_textless.png") as imgdata:
                    borders["token_textless"][color] = Image.open(imgdata)
                    borders["token_textless"][color].load()
            except KeyError:
                print(f"Token border for {color} not found")
            with archive.open(f"borders/token/{color}.png") as imgdata:
                borders["token"][color] = Image.open(imgdata)
                borders["token"][color].load()
            with archive.open(f"borders/{color}.png") as imgdata:
                backs[color] = Image.open(imgdata)
                backs[color].load()
            with archive.open(f"borders/default/{color}_box.png") as imgdata:
                borders["default"][f"{color}_box"] = Image.open(imgdata)
                borders["default"][f"{color}_box"].load()
            with archive.open(f"borders/default/{color}_mdfc.png") as imgdata:
                borders["default"][f"{color}_mdfc"] = Image.open(imgdata)
                borders["default"][f"{color}_mdfc"].load()

            with archive.open(f"borders/default/{color}_adv_left.png") as imgdata:
                borders["default"][f"{color}_adv_left"] = Image.open(imgdata)
                borders["default"][f"{color}_adv_left"].load()
            with archive.open(f"borders/default/{color}_adv_right.png") as imgdata:
                borders["default"][f"{color}_adv_right"] = Image.open(imgdata)
                borders["default"][f"{color}_adv_right"].load()

        return cls(
            lands=lands,
            saga=borders["saga"],
            default=borders["default"],
            token=borders["token"],
            token_textless=borders["token_textless"],
            backs=backs,
        )

    # def multi_land(self, colors: str) -> Image.Image:
    #     print(f"Multi colored land ({len(colors)})")
    #     return Image.open(get_border_resource("l.png"))
    #
    #
    # def default_land(style: Style) -> Image.Image:
    #     print("Default land")
    #     return Image.open(get_border_resource("l.png"))

    def merge_border_colors(self, colors: str) -> Image.Image:
        print(f"Dual colored land ({colors})")
        left_box = self.lands[f"{colors[0].lower()}_box"]
        right_box = self.lands[f"{colors[1].lower()}_box"]

        box = merge(left_box, right_box)
        bg = self.lands["l_back"]
        lines = self.lands["l_lines"]

        bg.alpha_composite(lines)
        bg.alpha_composite(box)
        return bg

    def adjust_crop(self, bbox: BBox, border_size: int) -> BBox:
        return (
            bbox[0] + border_size,
            bbox[1] + border_size,
            bbox[2] - border_size,
            bbox[3] - border_size,
        )

    def merge_border_colors_dual(self, colors: str) -> Image.Image:
        print(f"Dual colored land ({colors}) OG Dual")
        left_box = self.lands[f"{colors[0].lower()}_box"]
        right_box = self.lands[f"{colors[1].lower()}_box"]

        left_bbox = left_box.getbbox(alpha_only=True)
        if left_bbox is None:
            left_bbox = (0, 0, left_box.width, left_box.height)

        box = merge(left_box, right_box)
        border_size = 20
        slice_size = 34
        boxes = [right_box, left_box]
        iterations = 8
        for i in range(iterations):
            box.alpha_composite(
                boxes[i % 2].crop(self.adjust_crop(left_bbox, border_size)),
                (
                    left_bbox[0] + border_size,
                    left_bbox[1] + border_size,
                ),
            )
            border_size += slice_size

        bg = self.lands["l_back"]
        lines = self.lands["l_lines"]

        bg.alpha_composite(lines)
        bg.alpha_composite(box)
        return bg

    def single_color_land(self, color: str) -> Image.Image:
        print(f"Sinlgle colored land ({color})")
        box = self.lands[f"{color.lower()}_box"]

        bg = self.lands["l_back"]
        lines = self.lands["l_lines"]

        bg.alpha_composite(lines)
        bg.alpha_composite(box)
        return bg

    def get_land_border(self, face: Face) -> Image.Image:
        colors = face.border_color
        if colors is None or len(colors) > 2 or colors.lower() == "c":
            return self.single_color_land("l")
        elif len(colors) == 2:
            if face.name in [
                "Badlands",
                "Bayou",
                "Plateau",
                "Savannah",
                "Scrubland",
                "Taiga",
                "Tropical Island",
                "Tundra",
                "Underground Sea",
                "Volcanic Island",
            ]:
                if face.name in ["Bayou", "Plateau", "Savannah", "Taiga", "Thundra"]:
                    colors = colors[::-1]
                return self.merge_border_colors_dual(colors)
            return self.merge_border_colors(colors)
        elif len(colors) == 1:
            return self.single_color_land(colors)
        return self.single_color_land("l")

    # def get_artifact_border(self, face: Face) -> Image.Image:
    #     if face.border_color.lower() == "c":
    #         return self.default["a"]
    #     else:
    #         color_border = self.get_default_border(face)
    #         artifact_back = self.default["a"]
    #         artifact_box = self.default["a_box"]
    #
    #         back = overlay(artifact_back, color_border)
    #         back.alpha_composite(artifact_box)
    #         return back

    def merge_hybrid_saga(self, colors: str) -> Image.Image:
        print(f"Dual colored hybrid card ({colors})")
        left = self.saga[colors[0].lower()]
        right = self.saga[colors[1].lower()]

        return merge(left, right)

    def merge_hybrid(self, colors: list[str]) -> Image.Image:
        left = self.get_back(colors[0], False)
        right = self.get_back(colors[1], False)

        return merge(left, right)

    def get_plain_border(self, face: Face) -> Image.Image:
        colors = "".join(face.border_color)
        if len(colors) == 0:
            return self.default["c_back"]
        if len(colors) == 1:
            return self.default[f"{colors.lower()}_back"]
        if len(colors) == 2 and "/" in face.mana_cost or face.mana_cost == "":
            return self.merge_hybrid([f"{color}_back" for color in colors])
        else:
            return self.default["m_back"]

    def get_single_color(self, face: Face) -> str:
        colors = "".join(face.border_color)
        if len(colors) == 0:
            return "c"
        if len(colors) == 1:
            return colors.lower()
        if len(colors) == 2 and "/" in face.mana_cost or face.mana_cost == "":
            return f"h{colors.lower()}"
        else:
            return "m"

    def get_back(self, color: str, is_artifact: bool) -> Image.Image:
        if color[0] == "h":
            left = self.get_back(color[1], False)
            right = self.get_back(color[2], False)
            back = merge(left, right)
        else:
            back = self.backs[color]

        if is_artifact:
            if color == "c":
                back = self.backs["a"]
            else:
                back = overlay(self.backs["a"], back)
        return back

    def get_default_box(
        self, color: str, is_artifact: bool, is_leveler: bool
    ) -> Image.Image:
        if is_artifact:
            box = self.default["a_box"].copy()
        elif color[0] == "h":
            left = self.get_default_box(color[1], is_artifact, is_leveler)
            right = self.get_default_box(color[2], is_artifact, is_leveler)
            box = merge(left, right)
        else:
            box = self.default[f"{color}_box"].copy()

        if is_leveler:
            box_as_mask = ImageOps.invert(box.getchannel("A"))
            empty_image = Image.new("RGBA", box.size, (0, 0, 0, 0))
            shadow_image = self.default["leveler_shadow"]
            shadow = Image.composite(empty_image, shadow_image, box_as_mask)

            box.alpha_composite(shadow)
        return box

    def get_adventure_box(
        self, color_left: str, color_right: str, is_artifact: bool
    ) -> Image.Image:
        left_box = self.get_adventure_side(color_left, "left", False).copy()
        right_box = self.get_adventure_side(color_right, "right", is_artifact)

        left_box.alpha_composite(right_box)
        return left_box

    def get_default_border(self, face: Face) -> Image.Image:
        color = self.get_single_color(face)
        is_artifact = "Artifact" in face.types
        is_leveler = "leveler" == face.layout

        back = self.get_back(color, is_artifact).copy()
        if is_artifact:
            color = "a"

        box = self.get_default_box(color, is_artifact, is_leveler)
        outline = self.default["outline"]
        back.alpha_composite(outline)
        back.alpha_composite(box)
        return back

    def get_color_token_border(
        self, color: str, is_artifact: bool, is_textless: bool
    ) -> Image.Image:
        borders = self.token_textless if is_textless else self.token

        if color[0] == "h":
            left = self.get_color_token_border(color[1], False, is_textless)
            right = self.get_color_token_border(color[2], False, is_textless)
            border = merge(left, right)
        else:
            border = borders[color]

        if is_artifact:
            if color == "c":
                border = borders["a"]
            else:
                border = overlay(borders["a"], border)
        return border

    def get_token_border(self, face: Face) -> Image.Image:
        if "Land" in face.types:
            color = "l"
        else:
            color = self.get_single_color(face)
        is_artifact = "Artifact" in face.types
        return self.get_color_token_border(color, is_artifact, face.textless)

    def get_mdfc_badge(self, back: Face) -> Image.Image:
        if "Land" in back.types:
            colors = back.produces if back.produces is not None else []
        else:
            colors = back.colors
        if len(colors) > 1:
            color = "M"
        else:
            color = colors[0]

        if "Artifact" in back.types:
            if color == "C":
                return self.default["a_mdfc"]
            else:
                color_mdfc = self.default[f"{color.lower()}_mdfc"]
                artifact_mdfc = self.default["a_mdfc"]
                return overlay(artifact_mdfc, color_mdfc)

        return self.default[f"{color.lower()}_mdfc"]

    def get_adventure_side(
        self, color: str, side: Literal["left", "right"], is_artifact
    ) -> Image.Image:
        if is_artifact:
            return self.default[f"a_adv_{side}"]
        if color[0] == "h":
            left = self.get_adventure_side(color[1], side, is_artifact)
            right = self.get_adventure_side(color[2], side, is_artifact)
            return merge(left, right)
        return self.default[f"{color.lower()}_adv_{side}"]

    def get_saga_border(self, face: Face) -> Image.Image:
        colors = "".join(face.border_color)
        if "Land" in face.types:
            colors = "l"
        if len(colors) == 0:
            colors = "a"
        if len(colors) == 1:
            return self.saga[colors.lower()]
        if len(colors) == 2 and "/" in face.mana_cost or face.mana_cost == "":
            return self.merge_hybrid_saga(colors)
        else:
            return self.saga["m"]

    def get_adventure_border(self, face: Face, back: Face) -> Image.Image:
        color_right = self.get_single_color(face)
        color_left = self.get_single_color(back)
        is_artifact = "Artifact" in face.types

        background = self.get_back(color_right, is_artifact).copy()
        if is_artifact:
            color_right = "a"

        box = self.get_adventure_box(color_left, color_right, is_artifact)
        outline = self.default["outline"]
        background.alpha_composite(outline)
        background.alpha_composite(box)
        return background

    def get_border(
        self, card: Card, face: Face, back: Optional[Face] = None
    ) -> Optional[Image.Image]:
        if "Saga" in face.subtypes:
            border = self.get_saga_border(face).copy()
        elif card.layout == "adventure" and back is not None:
            border = self.get_adventure_border(face, back).copy()
        # elif card.layout == "leveler":
        #     border = self.get_leveler_border(face).copy()
        elif card.layout in [
            "token",
            "emblem",
            "double_faced_token",
        ] or face.layout in ["token", "emblem", "double_faced_token"]:
            border = self.get_token_border(face).copy()
        elif "Land" in face.types:
            border = self.get_land_border(face).copy()
        else:
            border = self.get_default_border(face).copy()

        if card.layout == "modal_dfc" and back is not None:
            border.alpha_composite(self.get_mdfc_badge(back))

        return border


def main() -> None:
    art1 = ArtInfo(None)
    art2 = ArtInfo(None)
    card = get_card(sys.argv[1], art1, art2)
    with ZipFile("resource_pack.zip") as archive:
        with archive.open("style.json") as stylefile:
            style = load(stylefile)

        borders = Borders.load(archive, style)
    back = card.faces[1] if len(card.faces) > 1 else None
    img = borders.get_border(card, card.faces[0], back)
    if img is not None:
        img.show()


if __name__ == "__main__":
    main()
