from collections.abc import Sequence
import sys


from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import A4
from PIL import Image


def mkpdf(image_paths: Sequence[str | Image.Image], output_path: str) -> None:
    # A4 page size in points
    page_width, page_height = A4

    # Magic: The Gathering card dimensions in points (1 inch = 72 points)
    card_width = 2.5 * 72  # 2.5 inches wide
    card_height = 3.5 * 72  # 3.5 inches tall

    # Small gap between cards
    gap = 0.5  # Thin line between images in points

    # Calculate grid layout to center on A4 page
    rows, cols = 3, 3
    total_width = cols * card_width + (cols - 1) * gap
    total_height = rows * card_height + (rows - 1) * gap

    # Offset to center the grid on the A4 page
    x_offset = (page_width - total_width) / 2
    y_offset = (page_height - total_height) / 2

    # Set up the canvas for the PDF
    pdf_canvas = canvas.Canvas(output_path, pagesize=A4)

    # Process images in pages of 9 (3x3 grid per page)
    for page_start in range(0, len(image_paths), rows * cols):
        # Select the batch of images for the current page
        page_images = image_paths[page_start : page_start + rows * cols]

        for i, image_path in enumerate(page_images):
            # Calculate grid position (row and column)
            row = i // cols
            col = i % cols
            x = x_offset + col * (card_width + gap)
            y = y_offset + (rows - row - 1) * (card_height + gap)

            # Open the image
            if isinstance(image_path, Image.Image):
                img = image_path
            else:
                img = Image.open(image_path)

            # Scale the image dimensions to fit the card size in the PDF
            img_width, img_height = img.size
            img_aspect_ratio = img_width / img_height
            card_aspect_ratio = card_width / card_height

            # Determine scaling to fit within the card dimensions
            if img_aspect_ratio > card_aspect_ratio:
                display_width = card_width
                display_height = card_width / img_aspect_ratio
            else:
                display_height = card_height
                display_width = card_height * img_aspect_ratio

            # Center the image within the grid cell
            x_centered = x + (card_width - display_width) / 2
            y_centered = y + (card_height - display_height) / 2

            # Draw the high-quality image directly onto the PDF canvas
            pdf_canvas.drawInlineImage(
                image_path,
                x_centered,
                y_centered,
                width=display_width,
                height=display_height,
            )

            # If we opened an image from a path, close it to free resources
            if not isinstance(image_path, Image.Image):
                img.close()

        # Finalize the page and add a new one if more images are left
        pdf_canvas.showPage()

    # Save the PDF
    pdf_canvas.save()


if __name__ == "__main__":
    mkpdf(sys.argv[2:], sys.argv[1])
