from dataclasses import dataclass
from itertools import zip_longest
from typing import Literal, TypeAlias
from PIL.ImageFont import FreeTypeFont

from .render import Line, Word, Chunk, Formatting

Alignment: TypeAlias = Literal["left"] | Literal["center"] | Literal["right"]


@dataclass
class State:
    is_italics: bool
    is_image: bool


def get_next_state(char: str, next_char: str, state: State) -> State:
    italics = (
        state.is_italics and char != ")" and next_char != "—"
    ) or next_char == "("
    image = (state.is_image and char != "}") or next_char == "{"

    if next_char == "*":
        italics = not italics
    return State(italics, image)


def make_chunk(
    text: str,
    state: State,
    font: FreeTypeFont,
    font_i: FreeTypeFont,
    alignment: Alignment,
    parskip: float,
    phantom_indent: str,
    color: tuple[int, int, int],
) -> Chunk:
    return Chunk(
        Formatting(
            font_i if state.is_italics else font,
            state.is_image,
            scale=0.8 if state.is_image else 1.0,
            color=color,
            alignment=alignment,
            parskip=parskip,
            offset_perc=0.1 if state.is_image else 0.0,
            phantom_indent=phantom_indent,
        ),
        text,
    )


def parse(
    text: str,
    font: FreeTypeFont,
    font_i: FreeTypeFont,
    parskip: float = 1.1,
    alignment: Alignment = "left",
    start_italics=False,
    phantom_indent="",
    color: tuple[int, int, int] = (0, 0, 0),
) -> Line:
    state = State(start_italics or text[0] == "(", text[0] == "{")
    next_state = State(False, False)
    chunk_str: str = ""
    word: list[Chunk] = []
    line: list[Word] = []

    if text[0] == "•":
        phantom_indent = "• "
        parskip = 1.0

    if "—" in text and text[0] != "—":
        state.is_italics = True
        if (
            "choose" in text
            or "Choose" in text
            or "Ward—" in text
            or "Impending" in text
            or "Escalate" in text
            or "Companion" in text
            or "Boast" in text
            or "Escape" in text
            or "Evoke" in text
            or "Exhaust" in text
            or "Prototype" in text
        ):
            state.is_italics = False

    for char, next_char in zip_longest(text, text[1:]):
        next_state = get_next_state(char, next_char, state)
        if char == " ":
            word.append(
                make_chunk(
                    chunk_str,
                    state,
                    font,
                    font_i,
                    alignment,
                    parskip,
                    phantom_indent,
                    color,
                )
            )
            chunk_str = ""
            line.append(word)
            word = []
        else:
            chunk_str += char if char not in "{}*" else ""
            if state != next_state or next_char == "{":
                word.append(
                    make_chunk(
                        chunk_str,
                        state,
                        font,
                        font_i,
                        alignment,
                        parskip,
                        phantom_indent,
                        color,
                    )
                )
                chunk_str = ""
        state = next_state
    if chunk_str:
        word.append(
            make_chunk(
                chunk_str,
                next_state,
                font,
                font_i,
                alignment,
                parskip,
                phantom_indent,
                color,
            )
        )
    if word:
        line.append(word)
    return line
