from dataclasses import dataclass, field
from typing import Literal, Optional, TypeAlias
from PIL import Image, ImageDraw, ImageFont
import math

from .resources import get_symbol_resource


class BoundingBoxExceeded(Exception):
    pass


@dataclass
class Formatting:
    font: ImageFont.FreeTypeFont
    is_image: bool = field(default=False)
    scale: float = field(default=1.0)
    color: tuple[int, int, int] = field(default=(0, 0, 0))
    alignment: Literal["left"] | Literal["center"] | Literal["right"] = field(
        default="center"
    )
    parskip: float = field(default=1.25)
    offset_perc: float = 0.0
    phantom_indent: str = ""

    def get_font(self, font_offset: int) -> ImageFont.FreeTypeFont:
        return self.font.font_variant(
            size=int((self.font.size - font_offset) * self.scale)
        )


@dataclass
class TypesetMeta:
    top: int
    bot: int
    skip: float
    font_size: int
    space_length: int
    alignment: Literal["left"] | Literal["center"] | Literal["right"]
    parskip: float
    indent: int
    empty_space_perc: float = field(default=0)


@dataclass
class Chunk:
    format: Formatting
    text: str


Word: TypeAlias = list[Chunk]
Line: TypeAlias = list[Word]


def image_name_for_symbol(symbol: str) -> str:
    stripped_name = symbol.replace("/", "").lower()
    match stripped_name:
        case "rb":
            return "br"
    return stripped_name


def render_chunk(
    chunk: Chunk, font_offset: int, symbols: dict[str, Image.Image]
) -> tuple[Image.Image, TypesetMeta]:
    font = chunk.format.get_font(font_offset)
    space_length = math.ceil(font.getlength(" "))
    indent = (
        font.getlength(chunk.format.phantom_indent)
        if chunk.format.phantom_indent
        else 0
    )

    if chunk.format.is_image:
        # symbol_full = Image.open(get_symbol_resource(f"{image_name_for_symbol(chunk.text)}.png"))
        symbol_full = symbols[image_name_for_symbol(chunk.text)]
        width = int(symbol_full.width / symbol_full.height * font.size)
        symbol = symbol_full.resize((width, font.size))
        image = Image.new(
            "RGBA", (width + 2 * (space_length // 8), int(font.size)), (0, 0, 0, 0)
        )
        image.paste(symbol, (space_length // 8, 0))
        return (
            image,
            TypesetMeta(
                top=int(font.size * (1 - chunk.format.offset_perc)),
                bot=int(font.size * chunk.format.offset_perc),
                skip=image.width,
                font_size=int(font.size),
                space_length=space_length,
                alignment=chunk.format.alignment,
                parskip=chunk.format.parskip,
                indent=int(indent),
            ),
        )

    else:
        _, top_box, width, height = font.getbbox(chunk.text)
        skip = font.getlength(chunk.text)
        top, bot = font.getmetrics()
        # print(top, bot)
        if bot < 0:
            bot = -bot

        image = Image.new("RGBA", (width, top + bot), (0, 0, 0, 0))
        canvas = ImageDraw.Draw(image)
        canvas.text(
            (0, top), chunk.text, anchor="ls", font=font, fill=chunk.format.color
        )
        return (
            image,
            TypesetMeta(
                top=top,
                bot=bot,
                skip=skip,
                font_size=font.size,
                space_length=space_length,
                alignment=chunk.format.alignment,
                parskip=chunk.format.parskip,
                indent=int(indent),
            ),
        )


def render_word(
    word: Word, font_offset: int, symbols: dict[str, Image.Image]
) -> tuple[Image.Image, TypesetMeta]:
    rendered_chunks = [render_chunk(c, font_offset, symbols) for c in word]

    max_top = max(meta.top for _, meta in rendered_chunks)
    max_bot = max(meta.bot for _, meta in rendered_chunks)
    max_font_size = max(meta.font_size for _, meta in rendered_chunks)
    parskip = rendered_chunks[0][1].parskip
    alignment = rendered_chunks[0][1].alignment
    indent = rendered_chunks[0][1].indent

    width = sum(img.width for img, _ in rendered_chunks)
    height = max_top + max_bot

    image = Image.new("RGBA", (width, height), (0, 0, 0, 0))

    x = 0
    for chunk_img, meta in rendered_chunks:
        image.alpha_composite(chunk_img, (int(x), max_top - meta.top))
        x += meta.skip

    return (
        image,
        TypesetMeta(
            top=max_top,
            bot=max_bot,
            skip=x,
            font_size=max_font_size,
            space_length=rendered_chunks[-1][1].space_length,
            alignment=alignment,
            parskip=parskip,
            indent=indent,
        ),
    )


def render_line(
    line: Line,
    font_offset: int,
    max_width: int,
    first_line: bool,
    symbols: dict[str, Image.Image],
) -> tuple[Image.Image, Line, TypesetMeta]:
    rendered_words: list[tuple[Image.Image, TypesetMeta]] = []
    last_space_width = 0
    width = 0

    is_first_word = True
    last_word = None

    total_space_width = 0
    for w in line:
        w_img, meta = render_word(w, font_offset, symbols)

        if is_first_word and not first_line:
            width += meta.indent
            is_first_word = False
        if width + w_img.width >= max_width:
            last_word = w_img, meta
            break
        else:
            last_space_width = meta.space_length
            width = width + meta.skip + meta.space_length
            total_space_width += meta.space_length
            rendered_words.append((w_img, meta))

    micro_px = 0
    if last_word is not None:
        micro_max = int(last_word[1].font_size / 14)
        overshoot = last_word[0].width
        empty_space = max_width - width
        diff = overshoot - empty_space
        for micro_px in range(micro_max + 1):
            if diff < micro_px * len(rendered_words):
                # print(f"{micro_px=}")
                rendered_words.append(last_word)
                width = width + last_word[1].skip + last_space_width
                break

    width -= last_space_width

    if width <= 0:
        raise BoundingBoxExceeded

    max_top = max(meta.top for _, meta in rendered_words)
    max_bot = max(meta.bot for _, meta in rendered_words)
    max_font_size = max(meta.font_size for _, meta in rendered_words)
    alignment = rendered_words[0][1].alignment
    parskip = rendered_words[0][1].parskip
    indent = rendered_words[0][1].indent

    height = max_top + max_bot

    image = Image.new("RGBA", (int(width), height), (0, 0, 0, 0))

    x = 0 if first_line else indent
    for word_img, meta in rendered_words:
        image.paste(word_img, (int(x), max_top - meta.top))
        x += meta.skip + meta.space_length - micro_px

    # x -= space_length

    return (
        image,
        line[len(rendered_words) :],
        TypesetMeta(
            top=max_top,
            bot=max_bot,
            skip=x,
            font_size=max_font_size,
            space_length=last_space_width,
            alignment=alignment,
            parskip=parskip,
            indent=indent,
            empty_space_perc=(max_width - width) / max_width,
        ),
    )


def render_line_wrapped(
    line: Line,
    font_offset: int,
    max_width: int,
    max_empty_space_perc: float,
    symbols: dict[str, Image.Image],
) -> tuple[Image.Image, TypesetMeta]:
    height = 0
    lines: list[tuple[Image.Image, TypesetMeta]] = []
    first_line = True
    while line:
        img, line, meta = render_line(line, font_offset, max_width, first_line, symbols)
        lines.append((img, meta))
        first_line = False
        if line:  # Check if this was not the last line
            if meta.empty_space_perc > max_empty_space_perc:
                raise BoundingBoxExceeded

    height = sum(img.height for img, _ in lines)
    _, first_meta = lines[0]
    _, last_meta = lines[-1]
    height = -first_meta.font_size + first_meta.top
    for img, meta in lines:
        height += meta.font_size
    height += last_meta.bot

    image = Image.new("RGBA", (max_width, height), (0, 0, 0, 0))
    # canvas = ImageDraw.Draw(image)

    y = -first_meta.font_size + first_meta.top
    for rendered_line, meta in lines:
        match first_meta.alignment:
            case "left":
                x = 0
            case "right":
                x = max_width - rendered_line.width
            case "center":
                x = (max_width - rendered_line.width) // 2
        y += meta.font_size
        image.alpha_composite(rendered_line, (x, y - meta.top))
        # canvas.rectangle(
        #     (
        #         x,
        #         y - meta.top,
        #         x + rendered_line.width,
        #         y - meta.top + rendered_line.height,
        #     ),
        #     outline=(0, 0, 0),
        # )

    return image, first_meta


def render(
    text: list[Line],
    max_width: int,
    max_height: int,
    font_offset: int,
    fill: bool,
    symbols: dict[str, Image.Image],
) -> Image.Image:
    image = Image.new("RGBA", (max_width, max_height), (0, 0, 0, 0))

    max_empty_space_perc = 0.35 if fill else 0.2

    paragraphs = [
        render_line_wrapped(line, font_offset, max_width, max_empty_space_perc, symbols)
        for line in text
    ]

    space = int(
        (max_height - sum(img.height for img, _ in paragraphs)) / (len(paragraphs))
    )

    _, first_meta = paragraphs[0]

    y = (
        -int(first_meta.font_size * first_meta.parskip - first_meta.top)
        if not fill
        else -space
    )
    for line, (p_img, meta) in zip(text, paragraphs):
        y += int(meta.font_size * meta.parskip - meta.top) if not fill else space
        image.alpha_composite(p_img, (0, y))
        y += p_img.height

    if y > max_height or space <= 0:
        raise BoundingBoxExceeded

    # canvas = ImageDraw.Draw(image)
    # canvas.rectangle((0, 0, image.width - 1, y - 1), outline=(0, 255, 0))

    cropped_image = image.crop((0, 0, image.width, y))

    return cropped_image


def render_fit(
    text: list[Line],
    max_width: int,
    max_height: int,
    fill: bool,
    symbols: dict[str, Image.Image],
) -> Image.Image:
    offset = 0
    while True:
        try:
            img = render(text, max_width, max_height, offset, fill, symbols)
            break
        except BoundingBoxExceeded:
            offset += 1
            # print(f"resize: {offset}")

    return img


def render2(
    text: list[list[tuple[Formatting, str]]],
    max_width: int,
    max_height: int,
    parskip: float = 1.2,
    font_offset: int = 0,
):
    for paragraph in text:
        x = 0
        while paragraph:
            format, group = paragraph.pop()


def render_old(
    text: list[list[tuple[Formatting, str]]],
    max_width: int,
    max_height: int,
    parskip: float = 1.2,
    font_offset: int = 0,
):
    image = Image.new("RGBA", (max_width, max_height), (0, 0, 0, 0))
    canvas = ImageDraw.Draw(image)
    font = text[0][0][0].font
    last_text_fontsize = font.size

    y, _ = font.getmetrics()
    x = 0
    max_x = 0
    for paragraph in text:
        x = 0

        for format, group in paragraph:
            fontsize = int((format.font.size - font_offset) * format.scale)
            font = format.font.font_variant(size=fontsize)
            if format.is_image:
                spacing = font.getlength(" ") // 4
                icons = [
                    Image.open("symbols/{symbol.lower()}.png").resize(
                        (fontsize, fontsize)
                    )
                    for symbol in group.split()
                ]
                if x + fontsize * len(icons) + (len(icons) - 1) * spacing >= max_width:
                    max_x = max(max_x, x)
                    x = 0
                    y += last_text_fontsize

                for icon in icons:
                    image.alpha_composite(icon, (int(x), int(y - 0.9 * fontsize)))
                    x += fontsize + spacing
                x -= spacing
            else:
                last_text_fontsize = fontsize

                fits_in_line = group.split(" ")
                while fits_in_line:
                    does_not_fit = []
                    length = font.getlength(" ".join(fits_in_line))
                    while x + length >= max_width:
                        if not fits_in_line:
                            break
                        does_not_fit = [fits_in_line[-1]] + does_not_fit
                        fits_in_line = fits_in_line[:-1]
                        length = font.getlength(" ".join(fits_in_line))

                    canvas.text(
                        (x, y),
                        " ".join(fits_in_line),
                        anchor="ls",
                        font=font,
                        fill=format.color,
                    )
                    x += length
                    if does_not_fit:
                        max_x = max(max_x, x)
                        x = 0
                        y += fontsize
                    fits_in_line = does_not_fit
        y += int((font.size - font_offset) * parskip)
    _, bot = font.getmetrics()
    max_y = y + bot - (font.size - font_offset) * parskip

    if max_x > max_width or max_y > max_height:
        raise BoundingBoxExceeded
    return image, (max_x, max_y)


def test(mw: int):
    font = ImageFont.truetype("./mplantin.ttf", size=40)
    font2 = ImageFont.truetype("./mplantin-i.ttf", size=40)
    l: Line = [
        [Chunk(Formatting(font), "Test")],
        [
            Chunk(Formatting(font), "123"),
            Chunk(Formatting(font2), "("),
            Chunk(Formatting(font2, True, scale=0.8), "U"),
            Chunk(Formatting(font2, True, scale=0.8), "G"),
            Chunk(Formatting(font2, True, scale=0.8), "G"),
        ],
        [Chunk(Formatting(font2), "is")],
        [
            Chunk(Formatting(font2), "both)"),
            Chunk(Formatting(font, True, scale=0.8), "C"),
        ],
    ]

    return render_line_wrapped(l, 0, mw)


def main(max_width, max_height=1200, parskip=1.2):
    li = "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua."
    li2 = "At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.XXXg"
    fonti = ImageFont.truetype("./mplantin.ttf", size=28)
    li_split = li.split(" ")
    li2_split = li2.split(" ")
    li_line: Line = [
        [Chunk(Formatting(fonti, False, 1.0, (0, 0, 0)), w)] for w in li_split
    ]
    li2_line: Line = [
        [Chunk(Formatting(fonti, False, 1.0, (0, 0, 0)), w)] for w in li2_split
    ]
    font = ImageFont.truetype("./mplantin.ttf", size=40)
    font2 = ImageFont.truetype("./mplantin-i.ttf", size=40)

    format = Formatting(font, False, 1.0, (0, 0, 0))

    chunk = Chunk(format, "Test123")

    format2 = Formatting(font, False, 2.0, (0, 0, 0))

    chunk2 = Chunk(format2, "Test123")

    word1 = [chunk, chunk2]

    format3 = Formatting(font2, False, 1.4, (0, 255, 0))
    word2: Word = [Chunk(format3, "Blubbl"), (Chunk(format2, "Dal"))]

    # return render_line_wrapped([word1, word2], 0, max_width)
    return render_fit([li_line, li2_line], max_width, max_height, parskip)
