import re
from PIL import Image
from io import BytesIO
from typing import Any, Optional

from .datatypes import ArtInfo, Card, Face
from .mtgpics_overrides import art_override
from .art import find_artwork
from .cache_url import get


def get_card(
    query_or_cardjson: str | dict[str, Any], art1: ArtInfo, art2: ArtInfo
) -> Card:
    if isinstance(query_or_cardjson, dict):
        card_json = query_or_cardjson
    else:
        response = get(f"https://api.scryfall.com/cards/search?q={query_or_cardjson}")
        card_json = response.json()["data"][0]

    mtg_set = card_json["set"]
    cn = card_json["collector_number"]
    cn_int = int("".join(c for c in cn if c in "0123456789"))
    cn_add = "".join(c for c in cn if c not in "0123456789")

    rarity = card_json["rarity"][0]
    if rarity == "b":
        rarity = "m"
    layout = card_json["layout"]
    #
    # images: list[Optional[Image.Image]] = [None, None]
    if art1.url is None:
        if (mtg_set, cn_int) in art_override:
            art1.merge(art_override[(mtg_set, cn_int)][0])
        if art1.url is None:
            try:
                scryfall_art_url = card_json["image_uris"]["art_crop"]
            except KeyError:
                scryfall_art_url = card_json["card_faces"][0]["image_uris"]["art_crop"]
            art1.url = find_artwork(mtg_set, cn, scryfall_art_url)
    if art2.url is None:
        if (mtg_set, cn_int) in art_override:
            art2.merge(art_override[(mtg_set, cn_int)][1])
        if art2.url is None:
            try:
                scryfall_art_url = card_json["image_uris"]["art_crop"]
            except KeyError:
                scryfall_art_url = card_json["card_faces"][1]["image_uris"]["art_crop"]
            art2.url = find_artwork(mtg_set, cn, scryfall_art_url, True)

    if art1.url is not None:
        art1.image = Image.open(BytesIO(get(art1.url).content))
    else:
        raise RuntimeError(f"Could not find image for {mtg_set=} {cn=}")
    if art2.url is not None:
        art2.image = Image.open(BytesIO(get(art2.url).content))

    if layout in ["split"]:
        image_url = card_json["image_uris"]["art_crop"]
        r = get(image_url)
        image = Image.open(BytesIO(r.content))
        if art1.image is None:
            art1.image = image.crop((0, 0, image.width // 2, image.height))
        if art2.image is None:
            art2.image = image.crop((image.width // 2, 0, image.width, image.height))

    faces = []

    if "card_faces" in card_json:
        if layout == "adventure":
            produces = card_json.get("produced_mana")
            for card in card_json["card_faces"]:
                if "Land" in card["type_line"] and produces is not None:
                    card["produced_mana"] = produces[0]
                    produces = produces[1:]
                card["set"] = card_json["set"]
                card["collector_number"] = card_json["collector_number"]
                card["image_uris"] = card_json["image_uris"]
                faces.append(get_face(card, art1, layout))
        else:
            produces = card_json.get("produced_mana")
            for art, card in zip([art1, art2], card_json["card_faces"]):
                if "Land" in card["type_line"] and produces is not None:
                    card["produced_mana"] = guess_produces(card["oracle_text"])
                    # produces = produces[1:]
                card["set"] = card_json["set"]
                card["collector_number"] = card_json["collector_number"]
                if art.image is None:
                    art.image = Image.open(
                        BytesIO(get(card["image_uris"]["art_crop"]).content)
                    )

                faces.append(get_face(card, art, layout))
    else:
        faces = [get_face(card_json, art1, layout)]

    return Card(
        rarity=rarity,
        layout=layout,
        faces=faces,
        cn=cn,
        cn_int=cn_int,
        cn_add=cn_add,
        set=mtg_set,
    )


def guess_produces(oracle_text: str) -> list[str]:
    add_in_text = re.findall(r"[aA]dd ((?:{[WUBRGC]}\s?)*)", oracle_text)
    colors_in_text = [char for char in add_in_text[0] if char in "WUBRGC"]
    return colors_in_text


def get_super_type_sub(type_line: str) -> tuple[list[str], list[str], list[str]]:
    available_super_types = ["Legendary", "Basic", "Snow", "Ongoing", "World"]

    super_types: list[str] = []
    types: list[str] = []
    subtypes: list[str] = []

    if "—" in type_line:
        type_line, snd = type_line.split("—")
        subtypes = snd.split(" ")

    possible_types = type_line.split(" ")
    super_types = [ty for ty in possible_types if ty in available_super_types]
    types = [ty for ty in possible_types if ty not in available_super_types]

    return sorted(super_types), sorted(types), sorted(subtypes)


def canonically_sort_dual_colors(colors: str) -> str:
    colors = "".join(sorted(colors))
    match colors:
        case "BU":
            return "UB"
        case "BW":
            return "WB"
        case "GR":
            return "RG"
        case "RU":
            return "UR"
        case "UW":
            return "WU"
    return colors


def get_color(
    colors: Optional[list[str]], mana_cost: str, is_hybrid=False
) -> tuple[str, list[str]]:
    # print(colors, is_hybrid)
    if colors is not None:
        if len(colors) == 0:
            return "C", []
        if len(colors) == 1:
            return colors[0], [colors[0]]
        elif len(colors) == 2 and is_hybrid:
            return canonically_sort_dual_colors("".join(colors)), colors
        else:
            return "M", colors
    else:
        return get_color(
            list(set(m for m in mana_cost if m in "WUBRGC")), mana_cost, is_hybrid
        )


def is_fetchland(oracle_text: str) -> bool:
    return (
        "Search your library for" in oracle_text
        and "put it onto the battlefield" in oracle_text
    )


def get_colors_of_fetch(oracle_text: str) -> list[str]:
    basic_land_types = ["Plains", "Island", "Swamp", "Mountain", "Forest"]
    used_basic_land_types = filter(lambda typ: typ in oracle_text, basic_land_types)
    lands_to_colors = {
        "Plains": "W",
        "Island": "U",
        "Swamp": "B",
        "Mountain": "R",
        "Forest": "G",
    }

    colors = "".join(map(lambda land: lands_to_colors[land], used_basic_land_types))
    return list(canonically_sort_dual_colors(colors))


def get_face(card_json: dict, art: ArtInfo, layout: str) -> Face:
    supertypes, types, subtypes = get_super_type_sub(card_json["type_line"])
    produces: Optional[str] = (
        "".join(sorted(c for c in card_json["produced_mana"] if c != "C"))
        if "produced_mana" in card_json
        else None
    )

    colors = card_json.get("colors")
    if layout != "double_faced_token":
        layout = card_json["layout"] if "layout" in card_json else "default"
    else:
        layout = "token"

    if "Land" in types:
        if is_fetchland(card_json["oracle_text"]):
            colors = get_colors_of_fetch(card_json["oracle_text"])
            # print(f"Fetchland {colors}")
        else:
            colors = [m for m in produces] if produces is not None else []
            if len(colors) > 2:
                colors = None

    # print(f"{colors=}")

    border_color, colors = get_color(
        colors,
        card_json["mana_cost"],
        is_fetchland(card_json["oracle_text"])
        or "/" in card_json["mana_cost"]
        or card_json["mana_cost"] == "",
    )
    # print(f"{border_color=}")
    # if card_json["name"] in self.color_overwrites:
    #     border_color = self.color_overwrites[card_json["name"]]

    flavor = correct_quotes(card_json.get("flavor_text"))
    oracletext = correct_quotes(card_json["oracle_text"])
    color_indicator = (
        "".join(sorted(card_json["color_indicator"]))
        if "color_indicator" in card_json
        else None
    )

    textless = not (bool(oracletext) or bool(flavor))

    if oracletext is None:
        oracletext = ""

    if layout == "leveler":
        oracle_lines = oracletext.split("\n")
        oracletexts = [oracle_lines[0]]
        for line in oracle_lines[1:]:
            if line.startswith("LEVEL "):
                oracletexts.append(line)
            else:
                oracletexts[-1] += f"\n{line}"
    else:
        oracletexts = [oracletext]

    # Stupid font
    artist = card_json["artist"].replace("ǵ", "g")

    # Some special overrides
    if "Dungeon" in types:
        oracletexts = [""]
        card_json["name"] = card_json["type_line"]
        card_json["type_line"] = "Card"
        textless = True

    if card_json["name"] == "Energy Reserve":
        textless = True
        oracletexts = [""]

    if card_json["name"] in ["Poison Counter", "Experience"]:
        textless = True
        layout = "token"
        oracletexts = [""]

    return Face(
        name=(card_json["name"]),
        mana_cost=card_json["mana_cost"],
        colors=colors,
        type_line=(
            card_json["type_line"]
            # if card_json["type_line"] != "Card"
            # else f"{card_json["name"]}"
        ),
        supertypes=supertypes,
        types=types,
        subtypes=subtypes,
        oracle_text=oracletexts,
        flavor_text=flavor,
        img=art,
        artist=artist,
        border_color=border_color,
        produces=produces,
        power=card_json.get("power"),
        toughness=card_json.get("toughness"),
        loyalty=card_json.get("loyalty"),
        color_indicator=color_indicator,
        layout=layout,
        textless=textless,
    )


def correct_quotes(text: str | None) -> Optional[str]:
    if text is None:
        return None
    quote = text

    while '"' in quote:
        quote = quote.replace('"', "“", 1).replace('"', "”", 1)
    quote = re.sub(r"'({)", r"‘\1", quote)
    quote = re.sub(r"'(\W)", r"’\1", quote)
    quote = re.sub(r"(\W)'", r"\1‘", quote)
    quote = quote.replace("'", "’")
    quote = quote.replace("−", "–")
    quote = quote.replace("ǵ", "g")  # not available in font

    return quote
