from __future__ import annotations

import io
from dataclasses import dataclass, field
from typing import Literal, Optional
from zipfile import ZipFile
from json import load

from PIL import Image, ImageDraw, ImageFont

from .makeborders import Borders
from .datatypes import Dimensions, Face, Card, BBox
from .render import (
    Chunk,
    Formatting,
    Line,
    Word,
    render,
    render_fit,
    render_word,
)
from .parser import parse


@dataclass(frozen=True)
class Style:
    # frames: dict[str, dict[str, dict[str, Frame]]]
    dimensions: dict[str, Dimensions]
    borders: Borders
    symbols: dict[str, Image.Image]
    artist_prefix: str
    artist_align: Literal["left"] | Literal["center"] | Literal["right"]
    line_sep: int
    mana_cost_size: int
    rules_font_path: str
    rules_font_size: int
    flavor_font_path: str
    header_font_path: str
    header_font_size: int
    type_font_size: int
    set_symbol_size: int
    artist_font_size: int
    mdfc_font_size: int
    pt_font_path: str
    pt_font_size: int
    copyright_font_size: int
    adventure_header_font_path: str
    adventure_header_font_size: int
    adventure_mana_cost_size: int
    adventure_type_font_size: int
    adventure_type_font_path: str
    proxy_note_text: str
    rules_font_data: bytes
    flavor_font_data: bytes
    header_font_data: bytes
    pt_font_data: bytes
    adventure_type_font_data: bytes
    adventure_header_font_data: bytes
    proxy_note_font_size: int
    border_size: BBox
    bleed_border_size: int
    add_bleed: bool = field(default=False)

    @classmethod
    def load(cls, zipfile: str, add_bleed: bool) -> Style:
        with ZipFile(zipfile) as archive:
            with archive.open("style.json") as stylefile:
                style = load(stylefile)
                default = style["dimensions"]["default"]
                dimensions = {
                    name: Dimensions.load(default | dimensions)
                    for name, dimensions in style["dimensions"].items()
                }

            borders = Borders.load(archive, style)

            symbol_list = [
                name[8:]
                for name in archive.namelist()
                if name.startswith("symbols") and name.endswith(".png")
            ]

            symbols = {}
            for symbol in symbol_list:
                with archive.open(f"symbols/{symbol}") as imgdata:
                    symbols[symbol[:-4]] = Image.open(imgdata)
                    symbols[symbol[:-4]].load()

            style |= {
                "dimensions": dimensions,
                "borders": borders,
                "symbols": symbols,
                "rules_font_data": archive.open(
                    f"fonts/{style['rules_font_path']}"
                ).read(),
                "flavor_font_data": archive.open(
                    f"fonts/{style['flavor_font_path']}"
                ).read(),
                "header_font_data": archive.open(
                    f"fonts/{style['header_font_path']}"
                ).read(),
                "pt_font_data": archive.open(f"fonts/{style['pt_font_path']}").read(),
                "adventure_type_font_data": archive.open(
                    f"fonts/{style['adventure_type_font_path']}"
                ).read(),
                "adventure_header_font_data": archive.open(
                    f"fonts/{style['adventure_header_font_path']}"
                ).read(),
            }
            if add_bleed:
                style["border_size"] = (style["bleed_border_size"],) * 4
            style["add_bleed"] = add_bleed

        return cls(**style)

    def render_leveler_box(
        self,
        levels: Optional[str],
        pt: str,
        abilities: str,
        bbox: BBox,
        font_size: Optional[int] = None,
        font_factor: float = 1,
    ) -> Image.Image:
        font_size = (
            font_size
            if font_size is not None
            else int(self.rules_font_size * font_factor)
        )
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data), size=font_size * font_factor
        )
        halffont = ImageFont.truetype(
            io.BytesIO(self.rules_font_data), size=font_size / 2 * font_factor
        )
        flavor_font = ImageFont.truetype(
            io.BytesIO(self.flavor_font_data), size=font_size
        )
        pt_font = ImageFont.truetype(
            io.BytesIO(self.pt_font_data), size=self.pt_font_size * font_factor
        )

        abilities_to_render: list[Line] = [
            parse(paragraph, font, flavor_font) for paragraph in abilities.splitlines()
        ]

        if levels is not None:
            _, _, ls_width, ls_height = halffont.getbbox("LEVEL")
            _, _, lwidth, lheight = font.getbbox(levels)
            levels_height = int(lheight + ls_height)
            levels_width = int(max(ls_width, lwidth))
            levels_img = Image.new(
                "RGBA",
                (
                    levels_width,
                    levels_height,
                ),
            )

            levels_canvas = ImageDraw.Draw(levels_img)
            levels_canvas.text((0, 0), "LEVEL", font=halffont, fill=(0, 0, 0))
            levels_canvas.text(
                ((levels_width - lwidth) // 2, int(ls_height)),
                levels,
                font=font,
                fill=(0, 0, 0),
            )
        else:
            levels_img = Image.new("RGBA", (0, 0))
            levels_height = 0

        if abilities_to_render:
            abilities_img = render_fit(
                abilities_to_render,
                bbox[2] * 4 // 5 if levels is None else bbox[2] * 3 // 5,
                bbox[3],
                False,
                self.symbols,
            )
        else:
            abilities_img = Image.new("RGBA", (0, 0))

        _, _, pt_width, pt_height = pt_font.getbbox(pt)
        pt_img = Image.new(
            "RGBA",
            (
                int(pt_width),
                int(pt_height),
            ),
        )
        pt_canvas = ImageDraw.Draw(pt_img)
        pt_canvas.text((0, 0), pt, font=font, fill=(0, 0, 0))

        height = max(levels_height, abilities_img.height, pt_height)

        img = Image.new(
            "RGBA",
            (
                bbox[2],
                int(height),
            ),
        )
        img.paste(levels_img, (0, int((height - levels_height) // 2)))
        img.paste(
            abilities_img,
            (
                0 if levels is None else (bbox[2] // 5),
                int((height - abilities_img.height) // 2),
            ),
        )
        img.paste(pt_img, (bbox[2] - int(pt_width), int((height - pt_height) // 2)))

        return img

    def render_rules_box(
        self,
        face: Face,
        shrink_rules: int,
        bbox: BBox,
        text: str,
        fontsize: Optional[int] = None,
        font_factor: float = 1,
        fit: bool = True,
    ) -> Image.Image:
        fontsize = (
            fontsize
            if fontsize is not None
            else int(self.rules_font_size * font_factor)
        )
        font = ImageFont.truetype(io.BytesIO(self.rules_font_data), size=fontsize)
        flavor_font = ImageFont.truetype(
            io.BytesIO(self.flavor_font_data), size=fontsize
        )

        text_to_render: list[Line] = [
            parse(paragraph, font, flavor_font) for paragraph in text.splitlines()
        ]
        if face.flavor_text is not None:
            text_to_render.extend(
                [
                    parse(
                        paragraph,
                        font,
                        flavor_font,
                        1 if paragraph[0] == "—" else 1.25,
                        "right" if paragraph[0] == "—" else "left",
                        start_italics=True,
                    )
                    for paragraph in face.flavor_text.splitlines()
                ]
            )

        if not text_to_render:
            text_to_render = [
                Line([Word([Chunk(Formatting(font, is_image=False), " ")])])
            ]

        if len(text_to_render) == 1 and len(text_to_render[0]) <= 8:
            text_to_render[0][0][0].format.alignment = "center"

        if fit:
            rules_img = render_fit(
                text_to_render,
                bbox[2],
                bbox[3] - shrink_rules,
                "Saga" in face.subtypes,
                self.symbols,
            )
        else:
            rules_img = render(
                text_to_render,
                bbox[2],
                bbox[3] - shrink_rules,
                0,
                "Saga" in face.subtypes,
                self.symbols,
            )

        return rules_img

    def render_title_box(
        self,
        face: Face,
        font_size_offset: int,
        bbox: BBox,
        font_data: Optional[bytes] = None,
        fontsize: Optional[int] = None,
        font_factor: float = 1,
    ) -> tuple[Image.Image, float]:
        title = face.name
        if face.layout in ["emblem"]:
            title = " ".join("emblem".upper())
        elif face.type_line == "Card":
            title = " ".join("marker".upper())
        elif face.layout in ["token", "double_faced_token"]:
            name_split = set(face.name.split(" "))
            subtypes_split = set(
                (subtype for subtype in face.subtypes if subtype != "")
            )
            if name_split == subtypes_split:
                title = " ".join("token".upper())
            else:
                title = " ".join(face.name.upper())

        font_data = font_data if font_data is not None else self.header_font_data
        fontsize = int(
            (
                (fontsize if fontsize is not None else self.header_font_size)
                - font_size_offset
            )
            * font_factor
        )

        font = ImageFont.truetype(
            io.BytesIO(font_data),
            size=fontsize,
        )
        _, _, width, _ = font.getbbox(title)
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        canvas = ImageDraw.Draw(img)
        canvas.text((3, 3), title, font=font, fill=(0, 0, 0))
        canvas.text((0, 0), title, font=font, fill=(255, 255, 255))
        return img, width + 3

    def render_mdfc_box_text_right(
        self, face: Face, bbox: BBox, font_size_offset: int
    ) -> tuple[Image.Image, int]:
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data),
            size=self.mdfc_font_size - font_size_offset,
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        text = [
            line for line in face.oracle_text[0].splitlines() if "{T}: Add" in line
        ][0]
        parsed_text = parse(
            text,
            font,
            font,
            color=(255, 255, 255) if face.colors != ["W"] else (0, 0, 0),
        )
        # print(face.oracle_text)
        text_img = render_fit(
            [parsed_text],
            bbox[2],
            bbox[3],
            False,
            self.symbols,
        )
        tbbox = text_img.getbbox()
        if tbbox is None:
            raise RuntimeError("No text on backside")
        _, _, width, height = tbbox
        # left, top, right, bottom = font.getbbox(text)
        y = (bbox[3] - height) // 2
        img.alpha_composite(text_img, (0, y))

        # canvas.text((3, y + 3), text, font=font, fill=(0, 0, 0))
        # canvas.text((0, y), text, font=font, fill=(255, 255, 255))
        return img, width

    def render_mdfc_box_text(
        self, face: Face, bbox: BBox, font_size_offset: int
    ) -> tuple[Image.Image, float]:
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data),
            size=self.mdfc_font_size - font_size_offset,
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        canvas = ImageDraw.Draw(img)
        text = (
            " ".join(face.subtypes)
            if "Creature" in face.types or "Planeswalker" in face.types
            else " ".join(face.types)
        )
        left, top, right, bottom = font.getbbox(text)
        y = (bbox[3] - bottom) // 2
        canvas.text((3, y + 3), text, font=font, fill=(0, 0, 0))
        canvas.text((0, y), text, font=font, fill=(255, 255, 255))
        return img, right + 3

    def render_type_box(
        self,
        face: Face,
        font_size_offset: int,
        bbox: BBox,
        font_data: Optional[bytes] = None,
        fontsize: Optional[int] = None,
        font_factor: float = 1,
    ) -> tuple[Image.Image, float]:
        font_data = font_data if font_data is not None else self.rules_font_data
        fontsize = (
            fontsize if fontsize is not None else int(self.type_font_size * font_factor)
        )
        font = ImageFont.truetype(
            io.BytesIO(font_data),
            size=fontsize - font_size_offset,
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )

        type_line = face.type_line
        if face.layout in ["emblem"] or face.type_line == "Card":
            type_line = face.name

        canvas = ImageDraw.Draw(img)
        left, top, right, bottom = font.getbbox(type_line)
        a, d = font.getmetrics()
        y = (bbox[3] - (a // 2)) // 4
        canvas.text((3, y + 3), type_line, font=font, fill=(0, 0, 0))
        canvas.text((0, y), type_line, font=font, fill=(255, 255, 255))
        return img, right + 3

    def render_artist_box(
        self, face: Face, bbox: BBox, font_factor: float = 1
    ) -> tuple[Image.Image, float]:
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data),
            size=int(self.artist_font_size * font_factor),
        )

        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        text = f"{self.artist_prefix}{face.artist}"

        canvas = ImageDraw.Draw(img)
        left, top, right, bottom = font.getbbox(text)
        canvas.text((3, 3), text, font=font, fill=(0, 0, 0))
        canvas.text((0, 0), text, font=font, fill=(255, 255, 255))
        return img, right + 3

    def render_proxy_box(
        self, face: Face, bbox: BBox, font_factor: float = 1
    ) -> Image.Image:
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data),
            size=int(self.proxy_note_font_size * font_factor),
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        text = self.proxy_note_text
        color = (
            (0, 0, 0)
            if face.colors == ["W"] and "Land" not in face.types
            else (255, 255, 255)
        )

        canvas = ImageDraw.Draw(img)
        left, top, right, bottom = font.getbbox(text)
        canvas.text((0, 0), text, font=font, fill=color)
        return img

    def render_copyright_box(
        self, face: Face, card: Card, bbox: BBox, font_factor: float = 1
    ) -> tuple[Image.Image, float]:
        font = ImageFont.truetype(
            io.BytesIO(self.rules_font_data),
            size=int(self.copyright_font_size * font_factor),
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        text = f"™ & © 2025 Wizards of the Coast {card.set.upper()} {card.cn}"

        color = (
            (0, 0, 0)
            if face.colors == ["W"] and "Land" not in face.types
            else (255, 255, 255)
        )

        canvas = ImageDraw.Draw(img)
        left, top, right, bottom = font.getbbox(text)
        canvas.text((0, 0), text, font=font, fill=color)
        return img, right

    def render_loyalty_box(
        self, face: Face, bbox: BBox, font_factor: float = 1
    ) -> Image.Image:
        font = ImageFont.truetype(
            io.BytesIO(self.pt_font_data), size=int(self.pt_font_size * font_factor)
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        canvas = ImageDraw.Draw(img)
        _, _, width, height = font.getbbox(f"{face.loyalty}")
        x = 200  # img.width - width
        if face.loyalty is None:
            raise RuntimeError("No loyalty for planeswalker")
        canvas.text((x, 4), face.loyalty, font=font, fill=(0, 0, 0))
        canvas.text(
            (x - 5, 0),
            face.loyalty,
            font=font,
            fill=(255, 255, 255),
        )
        return img

    def render_pt_box(
        self, face: Face, bbox: BBox, font_factor: float = 1
    ) -> Image.Image:
        font = ImageFont.truetype(
            io.BytesIO(self.pt_font_data), size=int(self.pt_font_size * font_factor)
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )
        canvas = ImageDraw.Draw(img)
        _, _, width, height = font.getbbox(f"{face.power}/{face.toughness}")
        x = img.width - width
        canvas.text((x, 4), f"{face.power}/{face.toughness}", font=font, fill=(0, 0, 0))
        canvas.text(
            (x - 5, 0),
            f"{face.power}/{face.toughness}",
            font=font,
            fill=(255, 255, 255),
        )
        return img

    def render_set_symbol(self, card: Card, bbox: BBox) -> tuple[Image.Image, int]:
        font = ImageFont.truetype(
            io.BytesIO(self.header_font_data), size=self.set_symbol_size
        )
        img = Image.new(
            "RGBA",
            (
                bbox[2],
                bbox[3],
            ),
        )

        line = [Chunk(Formatting(font, is_image=True), f"r{card.rarity}")]
        img, _ = render_word(line, 0, self.symbols)
        return img, img.width

    def render_mana_box(
        self, face: Face, bbox: BBox, fontsize: Optional[int] = None
    ) -> tuple[Image.Image, int]:
        if fontsize is None:
            fontsize = self.mana_cost_size

        font = ImageFont.truetype(io.BytesIO(self.header_font_data), size=fontsize)
        img = Image.new("RGBA", (bbox[2], bbox[3]))

        manacost = face.mana_cost[1:-1].split("}{")
        if manacost == [""]:
            return img, 0

        mana_render_line = [
            Chunk(Formatting(font, is_image=True, alignment="right"), mana)
            for mana in manacost
        ]

        img, _ = render_word(mana_render_line, 0, self.symbols)

        return img, img.width

    def render_card(self, card: Card) -> list[Image.Image]:
        if card.layout == "split":
            font_factor = 1.2

            # borders = (78, 81, 69, 82)
            # image_box = (borders[0], borders[1], 1500 - borders[2], 2100 - borders[3])
            # image_height = image_box[3] - image_box[1]
            # image_width = image_box[2] - image_box[0]
            #

            large_img_left = (
                self.render_face(
                    card.faces[0], card, font_factor=font_factor, add_border=False
                )
                # .crop(image_box)
                # .resize((new_width, new_height))
                # .transpose(Image.Transpose.ROTATE_90)
            )

            image_width, image_height = large_img_left.size
            factor = image_width / image_height
            new_height = int(image_height * factor)
            new_width = int(image_width * factor)

            img_left = large_img_left.resize((new_width, new_height)).transpose(
                Image.Transpose.ROTATE_90
            )

            img_right = (
                self.render_face(
                    card.faces[1], card, font_factor=font_factor, add_border=False
                )
                # .crop(image_box)
                .resize((new_width, new_height)).transpose(Image.Transpose.ROTATE_90)
            )

            complete_width = image_width + self.border_size[0] + self.border_size[2]
            complete_height = image_height + self.border_size[1] + self.border_size[3]

            img = Image.new("RGBA", (complete_width, complete_height), (0, 0, 0, 255))
            img.paste(
                img_left,
                (
                    self.border_size[0],
                    complete_height - new_width - self.border_size[3],
                    # 20 + new_width + int(2 * self.black_border_size * factor),
                ),
            )
            img.paste(img_right, (self.border_size[0], self.border_size[1] - 10))
            return [img]
        else:
            images = []
            if "adventure" in card.layout:
                images.append(self.render_face(card.faces[0], card, card.faces[1]))
            else:
                for facenr, face in enumerate(card.faces):
                    back = card.faces[1 - facenr] if len(card.faces) > 1 else None
                    images.append(self.render_face(face, card, back, facenr=facenr))
            return images

    def render_face(
        self,
        face: Face,
        card: Card,
        back: Optional[Face] = None,
        facenr: int = 0,
        font_factor: float = 1,
        add_border: bool = True,
    ) -> Image.Image:
        frame_type = "default"
        if "Saga" in face.subtypes:
            frame_type = "saga"
        elif "adventure" in card.layout:
            frame_type = "adventure"
        elif "leveler" in card.layout:
            frame_type = "leveler"
        elif "token" in card.layout or face.layout == "token":
            if face.textless:
                frame_type = "token_textless"
            else:
                frame_type = "token"
        elif "emblem" in card.layout or face.layout == "emblem":
            frame_type = "emblem"

        if face.img.image is None:
            raise RuntimeError(f"No image for face for card {face.name}")
        if face.img.zoom is None:
            face.img.zoom = 1.0

        is_transform = "transform" == card.layout

        card_img_full = face.img.image
        width_ratio = self.dimensions[frame_type].art_box[2] / card_img_full.width
        height_ratio = self.dimensions[frame_type].art_box[3] / card_img_full.height

        if height_ratio > width_ratio:
            art_height = self.dimensions[frame_type].art_box[3] * face.img.zoom
            art_width = card_img_full.width * height_ratio * face.img.zoom
        else:
            art_width = self.dimensions[frame_type].art_box[2] * face.img.zoom
            art_height = card_img_full.height * width_ratio * face.img.zoom

        resized_image = card_img_full.resize((int(art_width), int(art_height)))

        if face.img.x_perc_offset is None:
            face.img.x_perc_offset = 0.5
        if face.img.y_perc_offset is None:
            face.img.y_perc_offset = 0.5

        y_offset = int(
            (art_height - self.dimensions[frame_type].art_box[3])
            * face.img.y_perc_offset
        )
        x_offset = int(
            (art_width - self.dimensions[frame_type].art_box[2])
            * face.img.x_perc_offset
        )
        crop_box: BBox = (
            x_offset,
            y_offset,
            x_offset + self.dimensions[frame_type].art_box[2],
            y_offset + self.dimensions[frame_type].art_box[3],
        )

        card_img = resized_image.crop(crop_box)

        shrink_rules = 70 if "modal_dfc" in card.layout else 0

        border = self.borders.get_border(card, face, back)  # Image.open(border_path)
        if border is None:
            raise RuntimeError("")

        # Set Symbol
        if frame_type not in ["token", "emblem", "token_textless"]:
            set_img, set_width = self.render_set_symbol(
                card, bbox=self.dimensions[frame_type].set_symbol_box
            )
            border.alpha_composite(
                set_img,
                (
                    self.dimensions[frame_type].set_symbol_box[0]
                    + self.dimensions[frame_type].set_symbol_box[2]
                    - set_width,
                    self.dimensions[frame_type].set_symbol_box[1],
                ),
            )

        # Type line
        type_offset = 0
        box = self.dimensions[frame_type].type_box
        if face.color_indicator is not None:
            color_indicator_width = self.symbols[
                f"ci/{face.color_indicator.lower()}"
            ].width
            color_indicator_height = self.symbols[
                f"ci/{face.color_indicator.lower()}"
            ].height
            box = (box[0] - color_indicator_width // 2, box[1], box[2], box[3])
            border.alpha_composite(
                self.symbols[f"ci/{face.color_indicator.lower()}"],
                (box[0], box[1] + ((box[3] - color_indicator_height - 5) // 2)),
            )
            type_offset = color_indicator_width // 2 + 10
        type_size_offset = 0
        type_img, type_width = self.render_type_box(
            face,
            type_size_offset,
            bbox=box,
            font_factor=font_factor,
        )
        while type_width + type_offset > self.dimensions[frame_type].type_box[2]:
            type_size_offset += 1
            type_img, type_width = self.render_type_box(
                face,
                type_size_offset,
                box,
                font_factor=font_factor,
            )

        border.alpha_composite(
            type_img,
            (
                self.dimensions[frame_type].type_box[0] + type_offset,
                self.dimensions[frame_type].type_box[1],
            ),
        )

        if "adventure" in card.layout:
            if back is None:
                raise RuntimeError("Adventure card without back")

            adv_type_size_offset = 0
            adv_type_img, adv_type_width = self.render_type_box(
                back,
                adv_type_size_offset,
                self.dimensions[frame_type].type_box2,
                self.adventure_type_font_data,
                self.adventure_type_font_size,
                font_factor=font_factor,
            )
            while adv_type_width > self.dimensions[frame_type].type_box[2]:
                adv_type_size_offset += 1
                adv_type_img, adv_type_width = self.render_type_box(
                    face,
                    adv_type_size_offset,
                    self.dimensions[frame_type].type_box2,
                    self.adventure_type_font_data,
                    self.adventure_type_font_size,
                    font_factor=font_factor,
                )
            border.alpha_composite(
                adv_type_img,
                (
                    self.dimensions[frame_type].type_box2[0],
                    self.dimensions[frame_type].type_box2[1],
                ),
            )

        # Mana Box
        mana_img, mana_width = self.render_mana_box(
            face, bbox=self.dimensions[frame_type].mana_box
        )
        border.alpha_composite(
            mana_img,
            (
                self.dimensions[frame_type].mana_box[0]
                + self.dimensions[frame_type].mana_box[2]
                - mana_width,
                self.dimensions[frame_type].mana_box[1],
            ),
        )

        adv_mana_width = 0

        if "adventure" in card.layout:
            if back is None:
                raise RuntimeError("Adventure card without back")
            adv_mana_img, adv_mana_width = self.render_mana_box(
                back,
                self.dimensions[frame_type].mana_box2,
                self.adventure_mana_cost_size,
            )
            border.alpha_composite(
                adv_mana_img,
                (
                    self.dimensions[frame_type].mana_box2[0]
                    + self.dimensions[frame_type].mana_box2[2]
                    - adv_mana_width,
                    self.dimensions[frame_type].mana_box2[1],
                ),
            )

        # Title
        max_title_width = min(
            self.dimensions[frame_type].title_box[2],
            self.dimensions[frame_type].mana_box[2] - mana_width - 20,
        )
        title_size_offset = 0
        title_img, title_width = self.render_title_box(
            face,
            title_size_offset,
            bbox=self.dimensions[frame_type].title_box,
            font_factor=font_factor,
            font_data=(
                self.rules_font_data if face.layout in ["token", "emblem"] else None
            ),
        )
        while title_width >= max_title_width:
            # print(f"title resize: {title_size_offset}")
            title_size_offset += 1
            title_img, title_width = self.render_title_box(
                face,
                title_size_offset,
                bbox=self.dimensions[frame_type].title_box,
                font_factor=font_factor,
            )

        title_pos_x = self.dimensions[frame_type].title_box[0]
        if face.layout in ["token", "emblem"]:
            title_pos_x = int(
                self.dimensions[frame_type].title_box[0]
                + (max_title_width - title_width) // 2
            )

        border.alpha_composite(
            title_img,
            (
                # self.dimensions[frame_type].title_box[0],
                title_pos_x,
                self.dimensions[frame_type].title_box[1] + (title_size_offset // 2),
            ),
        )

        if is_transform:
            if facenr == 0:
                transform_icon = self.symbols["transform_front"]
                border.alpha_composite(
                    transform_icon,
                    (
                        self.dimensions[frame_type].title_left_box[0],
                        self.dimensions[frame_type].title_left_box[1],
                    ),
                )
            elif facenr == 1:
                transform_icon = self.symbols["transform_back"]
                border.alpha_composite(
                    transform_icon,
                    (
                        self.dimensions[frame_type].title_right_box[0],
                        self.dimensions[frame_type].title_right_box[1],
                    ),
                )

        if "adventure" in card.layout:
            if back is None:
                raise RuntimeError("Adventure card without back")
            max_adv_title_width = min(
                self.dimensions[frame_type].title_box2[2],
                self.dimensions[frame_type].mana_box2[2] - adv_mana_width - 20,
            )
            adv_title_size_offset = 0
            adv_title_img, adv_title_width = self.render_title_box(
                back,
                adv_title_size_offset,
                self.dimensions[frame_type].title_box2,
                self.adventure_header_font_data,
                self.adventure_header_font_size,
                font_factor=font_factor,
            )
            while adv_title_width >= max_adv_title_width:
                adv_title_size_offset += 1
                adv_title_img, adv_title_width = self.render_title_box(
                    back,
                    adv_title_size_offset,
                    self.dimensions[frame_type].title_box2,
                    self.adventure_header_font_data,
                    self.adventure_header_font_size,
                    font_factor=font_factor,
                )

            border.alpha_composite(
                adv_title_img,
                (
                    self.dimensions[frame_type].title_box2[0],
                    self.dimensions[frame_type].title_box2[1]
                    + (adv_title_size_offset // 2),
                ),
            )

        # Artist
        artist_img, artist_width = self.render_artist_box(
            face, bbox=self.dimensions[frame_type].artist_box, font_factor=font_factor
        )
        x = self.dimensions[frame_type].artist_box[0] + (
            (self.dimensions[frame_type].artist_box[2] - artist_width) // 2
            if self.artist_align == "center"
            else (
                self.dimensions[frame_type].artist_box[2] - artist_width
                if self.artist_align == "right"
                else 0
            )
        )
        border.alpha_composite(
            artist_img, (int(x), self.dimensions[frame_type].artist_box[1])
        )

        # Copyright
        copyright_img, copyright_width = self.render_copyright_box(
            face,
            card,
            bbox=self.dimensions[frame_type].copyright_box,
            font_factor=font_factor,
        )
        x = self.dimensions[frame_type].copyright_box[0] + (
            (self.dimensions[frame_type].copyright_box[2] - copyright_width) // 2
            if self.artist_align == "center"
            else (
                self.dimensions[frame_type].copyright_box[2] - copyright_width
                if self.artist_align == "right"
                else 0
            )
        )
        border.alpha_composite(
            copyright_img, (int(x), self.dimensions[frame_type].copyright_box[1])
        )

        # Proxy Note

        proxy_img = self.render_proxy_box(
            face, bbox=self.dimensions[frame_type].proxy_box, font_factor=font_factor
        )
        border.alpha_composite(
            proxy_img,
            (
                self.dimensions[frame_type].proxy_box[0],
                self.dimensions[frame_type].proxy_box[1],
            ),
        )

        # Rules

        if "leveler" in card.layout:
            rules_img = self.render_leveler_box(
                levels=None,
                pt=f"{face.power}/{face.toughness}",
                abilities=face.oracle_text[0],
                bbox=self.dimensions[frame_type].rules_boxes[0],
                font_factor=font_factor,
            )

            offset = (
                self.dimensions[frame_type].rules_boxes[0][3]
                - rules_img.height
                - shrink_rules
            ) // 2

            border.alpha_composite(
                rules_img,
                (
                    self.dimensions[frame_type].rules_boxes[0][0],
                    self.dimensions[frame_type].rules_boxes[0][1] + offset,
                ),
            )

            for i in range(1, 3):
                linesplit = face.oracle_text[i].split("\n")
                levels = linesplit[0][6:]
                pt = linesplit[1]
                abilities = "\n".join(linesplit[2:])

                rules_img = self.render_leveler_box(
                    levels=levels,
                    pt=pt,
                    abilities=abilities,
                    bbox=self.dimensions[frame_type].rules_boxes[i],
                    font_factor=font_factor,
                )
                offset = (
                    self.dimensions[frame_type].rules_boxes[i][3]
                    - rules_img.height
                    - shrink_rules
                ) // 2

                border.alpha_composite(
                    rules_img,
                    (
                        self.dimensions[frame_type].rules_boxes[i][0],
                        self.dimensions[frame_type].rules_boxes[i][1] + offset,
                    ),
                )
        else:
            rules_img = self.render_rules_box(
                face,
                shrink_rules,
                bbox=self.dimensions[frame_type].rules_boxes[0],
                text=face.oracle_text[0],
                font_factor=font_factor,
            )

            offset = (
                self.dimensions[frame_type].rules_boxes[0][3]
                - rules_img.height
                - shrink_rules
            ) // 2

            border.alpha_composite(
                rules_img,
                (
                    self.dimensions[frame_type].rules_boxes[0][0],
                    self.dimensions[frame_type].rules_boxes[0][1] + offset,
                ),
            )

        if "adventure" in card.layout:
            if back is None:
                raise RuntimeError("Adventure card without back")
            adv_rules_img = self.render_rules_box(
                back,
                shrink_rules,
                bbox=self.dimensions[frame_type].rules_boxes[1],
                text=back.oracle_text[0],
                fontsize=self.rules_font_size,
                font_factor=font_factor,
            )
            offset = (
                self.dimensions[frame_type].rules_boxes[1][3]
                - adv_rules_img.height
                - shrink_rules
            ) // 2
            border.alpha_composite(
                adv_rules_img,
                (
                    self.dimensions[frame_type].rules_boxes[1][0],
                    self.dimensions[frame_type].rules_boxes[1][1] + offset,
                ),
            )

        border.paste(
            card_img,
            (
                self.dimensions[frame_type].art_box[0],
                self.dimensions[frame_type].art_box[1],
            ),
        )

        # P/T

        if face.power is not None and frame_type != "leveler":
            border.alpha_composite(
                self.render_pt_box(
                    face,
                    bbox=self.dimensions[frame_type].pt_box,
                    font_factor=font_factor,
                ),
                (
                    self.dimensions[frame_type].pt_box[0],
                    self.dimensions[frame_type].pt_box[1],
                ),
            )

        # Loyality

        if face.loyalty is not None:
            border.alpha_composite(
                self.render_loyalty_box(
                    face,
                    bbox=self.dimensions[frame_type].pt_box,
                    font_factor=font_factor,
                ),
                (
                    self.dimensions[frame_type].pt_box[0],
                    self.dimensions[frame_type].pt_box[1] - 20,
                ),
            )

        # MDFC

        if "modal_dfc" in card.layout and back is not None:
            if "Land" not in back.types:
                mdfc_right, mdfc_right_width = self.render_mana_box(
                    back, self.dimensions[frame_type].mdfc_box, self.mdfc_font_size
                )
            else:
                mdfc_right, mdfc_right_width = self.render_mdfc_box_text_right(
                    back, self.dimensions[frame_type].mdfc_box, 0
                )

            border.alpha_composite(
                mdfc_right,
                (
                    self.dimensions[frame_type].mdfc_box[0]
                    + self.dimensions[frame_type].mdfc_box[2]
                    - mdfc_right_width,
                    self.dimensions[frame_type].mdfc_box[1] + 3,
                ),
            )

            mdfc_size_offset = 0
            mdfc_img, mdfc_width = self.render_mdfc_box_text(
                back,
                bbox=self.dimensions[frame_type].mdfc_box,
                font_size_offset=mdfc_size_offset,
            )
            while (
                mdfc_width
                > self.dimensions[frame_type].mdfc_box[2] - mdfc_right_width - 20
            ):
                mdfc_size_offset += 1
                mdfc_img, mdfc_width = self.render_mdfc_box_text(
                    face,
                    bbox=self.dimensions[frame_type].mdfc_box,
                    font_size_offset=mdfc_size_offset,
                )
            border.alpha_composite(
                mdfc_img,
                (
                    self.dimensions[frame_type].mdfc_box[0],
                    self.dimensions[frame_type].mdfc_box[1],
                ),
            )

        # self.draw_boxes(border, frame_type)
        if add_border:
            inner_width, inner_height = border.size

            img_width = inner_width + self.border_size[0] + self.border_size[2]
            img_height = inner_height + self.border_size[1] + self.border_size[3]

            img = Image.new("RGBA", (img_width, img_height), (0, 0, 0, 255))
            img.paste(border, (self.border_size[0], self.border_size[1]))

            # Draw all boxes

            return img
        else:
            return border

    def draw_boxes(self, img: Image.Image, frame_type: str) -> None:
        def absolute_box(t: tuple[int, int, int, int]) -> tuple[int, int, int, int]:
            return t[0], t[1], t[0] + t[2], t[1] + t[3]

        def zoom_box(
            t: tuple[int, int, int, int], z: float
        ) -> tuple[int, int, int, int]:
            x_offset = int(t[2] * (z - 1))
            y_offset = int(t[3] * (z - 1))
            return (t[0] + x_offset // 2, t[1], t[2] - x_offset, t[3] - y_offset)

        #
        canvas = ImageDraw.Draw(img)
        canvas.rectangle(
            absolute_box(zoom_box(self.dimensions[frame_type].art_box, 1.05)),
            outline=(255, 0, 0),
        )
        canvas.rectangle(
            absolute_box(zoom_box(self.dimensions[frame_type].art_box, 1.04)),
            outline=(0, 255, 0),
        )
        canvas.rectangle(
            absolute_box(zoom_box(self.dimensions[frame_type].art_box, 1.06)),
            outline=(0, 0, 255),
        )
        canvas.rectangle(
            absolute_box(zoom_box(self.dimensions[frame_type].art_box, 1.07)),
            outline=(0, 255, 0),
        )
        canvas.rectangle(
            absolute_box(zoom_box(self.dimensions[frame_type].art_box, 1.03)),
            outline=(0, 0, 255),
        )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].title_box),
        #     outline=(255, 0, 0),
        # )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].mana_box), outline=(255, 0, 0)
        # )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].type_box), outline=(255, 0, 0)
        # )
        # for box in self.dimensions[frame_type].rules_boxes:
        #     canvas.rectangle(
        #         absolute_box(box),
        #         outline=(255, 0, 0),
        #     )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].pt_box), outline=(255, 0, 0)
        # )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].set_symbol_box),
        #     outline=(255, 0, 0),
        # )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].artist_box),
        #     outline=(255, 0, 0),
        # )
        # canvas.rectangle(
        #     absolute_box(self.dimensions[frame_type].mdfc_box), outline=(255, 0, 0)
        # )
